#
# This file is part of VIRL 2
# Copyright (c) 2019-2025, Cisco Systems, Inc.
# All rights reserved.
#
from __future__ import annotations

import attr


@attr.s
class HostPort:
    host: str = attr.ib()
    port: int = attr.ib()
    is_ipv6: bool = attr.ib(default=False)

    @classmethod
    def parse(cls, hostname: str, port: int, force=True) -> HostPort:
        """Return hostport, using port from hostname if present and not forced"""
        is_ipv6 = False
        if hostname.startswith("["):
            index = hostname.find("]")
            if index > 0:
                is_ipv6 = True
                host = hostname[1:index]
                hostname = hostname[index:]
            else:
                raise ValueError("Missing right bracket for IPv6 address")

        index = hostname.rfind(":")
        if index > 0:
            if not force:
                port = int(hostname[index + 1 :])
            hostname = hostname[:index]
        if not is_ipv6:
            host = hostname
        return HostPort(host, port, is_ipv6)

    @property
    def url_host(self) -> str:
        if self.is_ipv6:
            return f"[{self.host}]"
        return self.host

    @property
    def url(self) -> str:
        return f"{self.url_host}:{self.port}"



#
# This file is part of VIRL 2
# Copyright (c) 2019-2025, Cisco Systems, Inc.
# All rights reserved.
#

from enum import StrEnum, auto


class BootEventType(StrEnum):
    MONITOR = "MONITOR"
    BOOTED = "BOOTED"


class DefaultPermissions:
    READ_ONLY = "read_only"
    READ_WRITE = "read_write"
    LAB_VIEW = "lab_view"
    LAB_EXEC = "lab_exec"
    LAB_EDIT = "lab_edit"
    LAB_ADMIN = "lab_admin"
    FULL_PERMISSIONS = [LAB_ADMIN, LAB_EXEC, LAB_EDIT, LAB_VIEW]
    VIEW_PERMISSIONS = [LAB_VIEW]
    FULL_PERMISSION_SET = set(FULL_PERMISSIONS)


class FSMState(StrEnum):
    pass


class LabState(FSMState):
    DEFINED_ON_CORE = "DEFINED_ON_CORE"
    STOPPED = "STOPPED"
    STARTED = "STARTED"


class NodeState(FSMState):
    DEFINED_ON_CORE = "DEFINED_ON_CORE"
    STOPPED = "STOPPED"
    STARTED = "STARTED"
    QUEUED = "QUEUED"
    BOOTED = "BOOTED"
    DISCONNECTED = "DISCONNECTED"

    @property
    def is_starting(self) -> bool:
        return (
            self is NodeState.QUEUED
            or self is NodeState.STARTED
            or self is NodeState.BOOTED
        )


class LinkState(FSMState):
    DEFINED_ON_CORE = "DEFINED_ON_CORE"
    STOPPED = "STOPPED"
    STARTED = "STARTED"


class InterfaceState(FSMState):
    DEFINED_ON_CORE = "DEFINED_ON_CORE"
    STOPPED = "STOPPED"
    STARTED = "STARTED"


class LabEventType(StrEnum):
    ADD = "ADD"
    REMOVE = "REMOVE"
    CHANGE = "CHANGE"
    STATE_CHANGED = "STATE_CHANGED"


class LabEventElementType(StrEnum):
    LAB = "LAB"
    NODE = "NODE"
    LINK = "LINK"
    INTERFACE = "INTERFACE"
    ANNOTATION = "ANNOTATION"
    CONNECTOR_MAPPING = "CONNECTOR_MAPPING"
    SMART_ANNOTATION = "SMART_ANNOTATION"


class OptInStatus(StrEnum):
    UNSET = auto()
    ACCEPTED = auto()
    DECLINED = auto()


class TelemetryEventCategory(StrEnum):
    START_LAB = auto()
    STOP_LAB = auto()
    QUEUE_NODE = auto()
    STOP_NODE = auto()
    START_NODE = auto()
    BOOT_NODE = auto()
    WIPE_NODE = auto()
    PACKET_CAPTURE = auto()
    LICENSE_INFO = auto()
    MAINTENANCE_STATE_CHANGE = auto()
    NOTICE_STATE_CHANGE = auto()
    RUNNING_NODES = auto()
    RESOURCE_POOL = auto()
    USER_ACTIVITY = auto()
    USER_GROUP = auto()
    SYSTEM_STATS = auto()
    BLKINFO = auto()
    VMWARE = auto()
    DMIINFO = auto()
    CPUINFO = auto()
    MEMINFO = auto()
    HYPERVISOR = auto()
    IMPORT_LAB = auto()
    EXPORT_LAB = auto()
    AAA_INFO = auto()



#
# This file is part of VIRL 2
# Copyright (c) 2019-2025, Cisco Systems, Inc.
# All rights reserved.
#
import re
from typing import Annotated, Literal, Union

from fastapi import Body
from pydantic import BaseModel, Field, model_serializer, model_validator

from simple_webserver.schemas.common import (
    DOMAIN_REG,
    PORT_REG,
    GroupName,
    Password,
    UserName,
    UUID4Type,
)


class UserAuthData(BaseModel, extra="forbid"):
    """Name and password of the user to be tested."""

    username: UserName = Field(...)
    password: Password = Field(...)


AuthDataBody = Annotated[
    UserAuthData,
    Body(
        description="This request body is a JSON object that holds authentication data."
    ),
]


CNArray = Annotated[
    list[str],
    Field(description="An array of LDAP CNs.", examples=["group-1", "group-2"]),
]

PublicCertificate = Annotated[
    str,
    Field(
        description="Contents of the public certificate in PEM format.",
        examples=[
            """
                -----BEGIN CERTIFICATE-----
                MIIDGzCCAgOgAwIBAgIBATANBgkq
                gno5gnopebgtAOFFHUnrr35n52/4
                //shortened
                xlJQaTOM9rpsuO/Q==
                -----END CERTIFICATE-----
                """
        ],
    ),
]

JwtToken = Annotated[
    str,
    Field(
        description="JWT token",
        pattern="^[A-Za-z0-9-_]+.[A-Za-z0-9-_]+.[A-Za-z0-9-_]+$",
        examples=[
            (
                "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJpc3MiOiJjb20uY2lzY28udmlybCIsI"
                "mlhdCI6MTYxMzc1MDgxNSwiZXhwIjoxNjEzODM3MjE1LCJzdWIiOiIwNDg5MDcxNS00YWE"
                "3LTRhNDAtYWQzZS1jZThmY2JkNGQ3YWEifQ.Q4heV5TTYQ6yhpJ5GKLm_Bf9D9NL-wDxL9"
                "Orz1ByxWs"
            )
        ],
    ),
]


class GroupAuthData(BaseModel, extra="forbid"):
    """Name of the group to be tested."""

    group_name: GroupName = Field(...)


class LocalAuthConfig(BaseModel, extra="forbid"):
    method: Literal["local"]


class LDAPAuthConfig(BaseModel, extra="forbid"):
    method: Literal["ldap"]
    manager_password: str | None = Field(
        default=None,
        max_length=256,
        description="""
          The password for the management user. If `use_ntlm` is `true` then the
          password will be converted to a NTLM hash and the hash is stored.
          Otherwise, the cleartext password will be stored using obfuscation.
        """,
    )
    server_urls: str | None = Field(
        default=None,
        max_length=256,
        description="URI of LDAP server, either LDAP or LDAPS, multiple servers can be specified, separate with space.",
        examples=["ldaps://ad.corp.com:3269"],
        pattern=re.compile(rf"^(?!.*[\n\r])(?:ldaps?://(?:{DOMAIN_REG})(?::{PORT_REG})?)(?: ldaps?://(?:{DOMAIN_REG})(?::{PORT_REG})?)*$"),
    )
    verify_tls: bool | None = Field(
        default=None,
        description="Set to `false` if certificates should not be verified.",
    )
    cert_data_pem: PublicCertificate | None = Field(default=None, description="Reference to a public certificate.")
    use_ntlm: bool | None = Field(
        default=None,
        description="If `true` then password for manager user will be stored as NTLM hash. Only works with ActiveDirectory servers.",
    )
    root_dn: str | None = Field(
        default=None,
        max_length=256,
        description="The root DN that will be applied.",
        examples=["DC=corp,DC=com"],
    )
    user_search_base: str | None = Field(
        default=None,
        max_length=256,
        description="The user search base where users should be looked up. Typically a OU or CN. Will be combined with the root DN.",
        examples=["CN=users,CN=accounts"],
    )
    user_search_filter: str | None = Field(
        default=None,
        max_length=1024,
        description="The filter that will be applied to the user. Must have a placeholder `{0}` replaced with the username.",
        examples=["(&(uid={0})(memberOf=CN=cmlusers,CN=groups,CN=accounts,DC=corp,DC=com))"],
    )
    admin_search_filter: str | None = Field(
        default=None,
        max_length=1024,
        description="Same as for the user search filter. Grants admin rights if matched.",
        examples=["(&(uid={0})(memberOf=CN=cmladmins,CN=groups,CN=accounts,DC=corp,DC=com))"],
    )
    group_search_base: str | None = Field(
        default=None,
        max_length=256,
        description="The group search base where groups should be looked up. Typically a OU or CN. Will be combined with the root DN.",
        examples=["CN=groups,CN=accounts"],
    )
    group_search_filter: str | None = Field(
        default=None,
        max_length=1024,
        description="The filter applied to groups. Must have a placeholder `{0}` replaced with the group name.",
        examples=["(&(cn={0})(objectclass=posixgroup))"],
    )
    group_via_user: bool | None = Field(
        default=None,
        description="If `true`, use `group_user_attribute` to determine user group memberships.",
    )
    group_user_attribute: str | None = Field(
        default=None,
        max_length=64,
        description="Attribute of the user that holds group memberships.",
        examples=["memberOf"],
    )
    group_membership_filter: str | None = Field(
        default=None,
        max_length=1024,
        description="Filter to apply to groups specifying the user.",
        examples=["(member={0})"],
    )
    manager_dn: str | None = Field(
        default=None,
        max_length=256,
        description="Manager user DN for lookup if anonymous search is not allowed.",
        examples=["uid=someuser,cn=users,cn=accounts,dc=corp,dc=com"],
    )
    display_attribute: str | None = Field(
        default=None,
        max_length=256,
        description="User attribute for displaying the logged in user.",
        examples=["displayName"],
    )
    group_display_attribute: str | None = Field(
        default=None,
        max_length=256,
        description="Group attribute for displaying group description.",
        examples=["description"],
    )
    email_address_attribute: str | None = Field(
        default=None,
        max_length=64,
        description="User attribute for displaying the email address.",
        examples=["mail"],
    )
    resource_pool: UUID4Type | None = Field(
        default=None, description="Resource pool or template ID for new user accounts."
    )


class RadiusAuthConfigBase(BaseModel, extra="forbid"):
    method: Literal["radius"]
    port: int = Field(
        default=1812,
        ge=1,
        le=65535,
        description="Default RADIUS server port (used when entry has no ':port').",
    )
    timeout: int = Field(
        default=5,
        ge=1,
        le=60,
        description="Timeout in seconds for RADIUS requests (default 5).",
    )
    nas_identifier: str | None = Field(
        default=None,
        max_length=256,
        description="NAS-Identifier to include in requests (optional).",
    )
    resource_pool: UUID4Type | None = Field(
        default=None,
        description="Resource pool or template ID for new user accounts.",
    )

    @classmethod
    def _validate_server_hosts_nullable(cls, server_hosts: str | None, allow_empty: bool) -> None:
        raw = (server_hosts or "").strip()
        if not raw:
            if allow_empty:
                return
            raise ValueError("server_hosts must contain at least one host")
        if any(ch in raw for ch in ("\n", "\r")):
            raise ValueError("server_hosts must not contain line breaks")
        for entry in raw.split():
            host, port_str = cls._split_host_port(entry)
            cls._validate_host_value(host)
            if port_str is not None:
                cls._validate_port_value(port_str)

    @staticmethod
    def _split_host_port(entry: str) -> tuple[str, str | None]:
        if ":" in entry:
            host, p = entry.rsplit(":", 1)
            return host, p
        return entry, None

    @staticmethod
    def _validate_host_value(host: str) -> None:
        if not re.match(rf"^{DOMAIN_REG}$", host):
            raise ValueError(f"Invalid host in server_hosts: {host}")

    @staticmethod
    def _validate_port_value(port_str: str) -> None:
        try:
            pi = int(port_str)
        except Exception:
            raise ValueError(f"Invalid port in server_hosts: {port_str}")
        if not (1 <= pi <= 65535):
            raise ValueError(f"Port out of range in server_hosts: {port_str}")


class RadiusAuthConfigRequest(RadiusAuthConfigBase):
    server_hosts: str = Field(
        ...,
        max_length=1024,
        description=(
            "Space-separated list of RADIUS servers. Each entry may be 'host' or 'host:port'. "
            "Entries without ':port' use the global 'port' value."
        ),
    )
    secret: str = Field(..., max_length=256, description="Shared secret for the RADIUS server(s).")

    @model_validator(mode="after")
    def validate_server_hosts(self):
        self.__class__._validate_server_hosts_nullable(self.server_hosts, allow_empty=False)
        return self


class RadiusAuthConfigResponse(RadiusAuthConfigBase):
    server_hosts: str | None = Field(
        default=None,
        max_length=1024,
        description=(
            "Space-separated list of RADIUS servers. Each entry may be 'host' or 'host:port'. "
            "Entries without ':port' use the global 'port' value."
        ),
    )
    secret: str | None = Field(
        default=None,
        max_length=256,
        description="Shared secret for the RADIUS server(s).",
    )

    @model_validator(mode="after")
    def validate_server_hosts(self):
        self.__class__._validate_server_hosts_nullable(self.server_hosts, allow_empty=True)
        return self


SystemAuthConfigRequest = Annotated[
    Union[LocalAuthConfig, LDAPAuthConfig, RadiusAuthConfigRequest],
    Field(discriminator="method"),
]
SystemAuthConfigResponse = Annotated[
    Union[LocalAuthConfig, LDAPAuthConfig, RadiusAuthConfigResponse],
    Field(discriminator="method"),
]


class SystemAuthConfigBase(BaseModel):
    config: SystemAuthConfigRequest

    @model_validator(mode="before")
    @classmethod
    def _wrap_union(cls, v):
        if isinstance(v, dict) and "config" not in v and v.get("method") is not None:
            return {"config": v}
        return v

    @model_serializer(mode="wrap")
    def _ser(self, handler):
        return handler(self.config)

    def __getattr__(self, item):
        return getattr(self.config, item)


class SystemAuthTestData(BaseModel, extra="forbid"):
    """
    Test Configuration Data which includes the system authentication
    configuration, and either a username and a password, or a group name.
    """

    auth_config: SystemAuthConfigBase
    auth_data: UserAuthData | None = None
    group_data: GroupAuthData | None = None

    @model_validator(mode="before")
    @classmethod
    def _map_aliases(cls, v):
        """
        normalize hyphenated input keys to field names.

        pydantic v2 emits warnings when using Field(alias=...) on union-typed
        fields. clients/tests send 'auth-config', 'auth-data', and
        'group-data'; we rewrite them to 'auth_config', 'auth_data', and
        'group_data' before validation to preserve input compatibility without
        triggering alias warnings.

        and then I wonder why we created hyphenated keys to begin with...
        But... need to maintain backwards-compatibility!
        """
        if isinstance(v, dict):
            if "auth-config" in v and "auth_config" not in v:
                v["auth_config"] = v.pop("auth-config")
            if "auth-data" in v and "auth_data" not in v:
                v["auth_data"] = v.pop("auth-data")
            if "group-data" in v and "group_data" not in v:
                v["group_data"] = v.pop("group-data")
        return v

    @model_validator(mode="after")
    def check_any_of(self):
        if self.auth_data is None and self.group_data is None:
            raise ValueError("Either auth_data or group_data must be provided.")
        return self


AuthTestAttributes = dict[str, str | list[str]]


class AuthTestUserResponse(BaseModel, extra="forbid"):
    auth_ok: bool | None = Field(
        default=None,
        description="The user has access to the system and the user filter matches.",
    )
    is_admin: bool | None = Field(
        default=None,
        description="The user has admin rights. If LDAP is configured, then the admin filter must match.",
    )
    display: str | None = Field(
        default=None,
        description="The user's display name, if the configured attribute was found.",
    )
    email: str | None = Field(
        default=None,
        description="The user's email address, if the configured attribute was found.",
    )
    attributes: AuthTestAttributes | None = Field(
        default=None,
        description=(
            "If the user auth'ed OK, then this object holds the dictionary of user attributes, "
            "independent of having CML access or admin access or other privileges. "
            "If an error occurred (config/server problem), then the error message will be included "
            "in the attributes."
        ),
    )


class AuthTestGroupResponse(BaseModel, extra="forbid"):
    group_ok: bool | None = Field(
        default=None,
        description="The group exists on the server and the group filter matches.",
    )
    is_member: bool | None = Field(
        default=None,
        description=(
            "Whether the given user is a member of the given group. Will default to false "
            "if either the user or the group could not be found on the LDAP server."
        ),
    )
    display: str | None = Field(
        default=None,
        description="The group's display name, if the configured attribute was found.",
    )
    attributes: AuthTestAttributes | None = Field(
        default=None,
        description=(
            "If the group was found OK, then this object holds the dictionary of group attributes, "
            "independent of having CML access or admin access or other privileges. "
            "If an error occurred (config/server problem), then the error message will be included "
            "in the attributes."
        ),
    )


class AuthTestResponse(BaseModel, extra="forbid"):
    """System Authentication Test response."""

    user: AuthTestUserResponse | None = Field(default=None, description="Results for the user.")
    group: AuthTestGroupResponse | None = Field(default=None, description="Results for the group.")


class AuthenticateResponse(BaseModel, extra="forbid"):
    """Authenticate response."""

    username: str | None = Field(default=None, examples=["admin"])
    id: UUID4Type | None = Field(default=None, description="ID of a user")
    token: JwtToken | None = Field(default=None)
    admin: bool | None = Field(default=None, examples=[False])
    error: str | None = Field(
        default=None,
        description="Error messages for errors that occurred while authenticating, "
        "but did not interrupt the login, such as LDAP group membership "
        "refresh errors.",
        examples=[
            "Could not refresh LDAP group memberships "
            "(Invalid base DN or root DN format)"
        ],
    )



#
# This file is part of VIRL 2
# Copyright (c) 2019-2025, Cisco Systems, Inc.
# All rights reserved.
#
from typing import Annotated, Literal

from pydantic import BaseModel, Field

from simple_webserver.schemas.common import (
    DeviceNature,
    DriverType,
    IPAddress,
    UUID4Type,
)
from simple_webserver.schemas.interfaces import InterfaceLabel
from simple_webserver.schemas.nodes import NodeLabel

PyAtsOs = Annotated[
    str,
    Field(
        description="The operating system as defined / understood by pyATS.",
        min_length=1,
        max_length=32,
        examples=["linux"],
    ),
]


PyAtsModel = Annotated[
    str,
    Field(
        description="The device model as defined by pyATS / Unicon.",
        min_length=1,
        max_length=32,
        examples=["cirrus"],
    ),
]

PyatsUsername = Annotated[
    str | None,
    Field(
        description="Username to use with pyATS / Unicon",
        min_length=1,
        max_length=64,
    ),
]

PyatsPassword = Annotated[
    str | None,
    Field(
        description="Password to use with pyATS / Unicon",
        min_length=1,
        max_length=128,
    ),
]


class PyAtsConnectionDetails(BaseModel, extra="forbid"):
    command: str | None = Field(default=None, examples=["open ..."])
    protocol: str = Field(..., examples=["ssh"])
    proxy: str | None = Field(default=None)
    ip: IPAddress | str | None = Field(
        default=None, description="IP address or hostname of the device"
    )
    port: int | None = Field(default=None, ge=0, le=65535)


class Connections(BaseModel, extra="forbid"):
    a: PyAtsConnectionDetails | None = Field(default=None)
    cli: PyAtsConnectionDetails | None = Field(default=None)
    defaults: dict[str, str] | None = Field(default=None)


class PyAtsCredentials(BaseModel, extra="forbid"):
    username: PyatsUsername = Field(default=None)
    password: PyatsPassword = Field(default=None)


class PyAtsDeviceCredentials(BaseModel, extra="forbid"):
    default: PyAtsCredentials


class PyAtsDevice(BaseModel, extra="forbid"):
    connections: Connections
    credentials: PyAtsDeviceCredentials
    os: PyAtsOs = Field(...)
    type: DeviceNature = Field(..., examples=["router"])
    model: PyAtsModel | None = Field(default=None)
    platform: DriverType | None = Field(default=None, examples=["iosv"])


class PyAtsInterfaceSummary(BaseModel, extra="forbid"):
    link: UUID4Type | None = Field(default=None, description="Link ID if present")
    type: Literal["ethernet", "loopback"] = Field(
        ..., description="interface type", examples=["ethernet"]
    )


class PyAtsDeviceTopology(BaseModel, extra="forbid"):
    interfaces: dict[InterfaceLabel, PyAtsInterfaceSummary]


class PyAtsTestbedName(BaseModel, extra="forbid"):
    name: str = Field(
        ...,
        min_length=1,
        max_length=64,
        description="The name of the testbed.",
        examples=["Lab at Thu 07:09 AM"],
    )


class PyAtsTestbed(BaseModel, extra="forbid"):
    devices: dict[NodeLabel, PyAtsDevice] = Field(..., description="Device list")
    testbed: PyAtsTestbedName = Field(..., description="Testbed details (name)")
    topology: dict[NodeLabel, PyAtsDeviceTopology] = Field(
        ..., description="Topology connections list"
    )



#
# This file is part of VIRL 2
# Copyright (c) 2019-2025, Cisco Systems, Inc.
# All rights reserved.
#
import re
from enum import StrEnum
from typing import Annotated, Literal

from pydantic import AfterValidator, BaseModel, Field, conlist, model_validator

from simple_webserver.schemas.common import (
    DefinitionID,
    DeviceNature,
    DriverType,
    FilePath,
)
from simple_webserver.schemas.nodes import NodeConfigurationContent, NodeParameters
from simple_webserver.schemas.pyats import PyAtsCredentials, PyAtsModel, PyAtsOs


class NicDriver(StrEnum):
    VIRTIO = "virtio"
    E1000 = "e1000"
    RTL8139 = "rtl8139"
    VMXNET3 = "vmxnet3"
    E1000E = "e1000e"
    E1000_82544GC = "e1000-82544gc"
    E1000_82545EM = "e1000-82545em"
    I82550 = "i82550"
    I82551 = "i82551"
    I82557A = "i82557a"
    I82557B = "i82557b"
    I82557C = "i82557c"
    I82558A = "i82558a"
    I82558B = "i82558b"
    I82559A = "i82559a"
    I82559B = "i82559b"
    I82559C = "i82559c"
    I82559ER = "i82559er"
    I82562 = "i82562"
    I82801 = "i82801"


class LibvirtDomainDriver(StrEnum):
    DOCKER = "docker"
    IOL = "iol"
    KVM = "kvm"
    NONE = "none"


class DiskDriver(StrEnum):
    IDE = "ide"
    SATA = "sata"
    VIRTIO = "virtio"


class VideoModel(StrEnum):
    VGA = "vga"
    CIRRUS = "cirrus"
    VMVGA = "vmvga"
    QXL = "qxl"
    XEN = "xen"
    VIRTIO = "virtio"
    NONE = "none"


class VideoDevice(BaseModel, extra="forbid"):
    memory: int = Field(..., description="Video Memory.", ge=1, le=128)
    model: VideoModel = Field(default=None, description="Video Model.")


class LinuxNative(BaseModel, extra="forbid"):
    """Base for simulation objects."""

    libvirt_domain_driver: LibvirtDomainDriver = Field(
        ..., description="Domain Driver."
    )
    driver: DriverType = Field(..., description="Simulation Driver.")
    disk_driver: DiskDriver = Field(default=None, description="Disk Driver.")
    efi_boot: bool = Field(default=None, description="If set, use EFI boot for the VM.")
    efi_code: FilePath = Field(default=None, description="EFI code file path; if unset, use default.")
    efi_vars: FilePath = Field(
        default=None,
        description="EFI NVRAM var template path; if unset, the code file "
        "is made writable; if set to constant 'stateless', "
        "the code file is marked stateless.",
    )
    machine_type: str = Field(
        default=None,
        description="QEMU machine type, defaults to pc; q35 is more modern.",
        min_length=1,
        max_length=32,
    )
    ram: int = Field(default=None, description="Memory in MiB.", ge=1, le=1048576)
    cpus: int = Field(default=None, description="CPUs.", ge=1, le=128)
    cpu_limit: int = Field(default=None, description="CPU Limit.", ge=20, le=100)
    cpu_model: str = Field(
        default=None,
        min_length=1,
        max_length=64,
        pattern=re.compile(r"^[a-zA-Z\d-]{1,32}(,[+!?^-][a-z\d._]{1,16})*(?![\n\r])$"),
    )
    nic_driver: NicDriver = Field(
        default=None,
        description="Network Driver.",
    )
    data_volume: int = Field(default=None, description="Data Disk Size in GiB.", ge=0, le=4096)
    boot_disk_size: int = Field(default=None, description="Boot Disk Size in GiB.", ge=0, le=4096)
    video: VideoDevice = Field(default=None, description="If present, then VNC can be used with the node VM.")
    enable_rng: bool = Field(
        default=True,
        description="If set, use a random number generator.",
    )
    enable_tpm: bool = Field(
        default=False,
        description="If set, enable an emulated TPM 2.0.",
    )

    @model_validator(mode="after")
    def validate(self):
        required: list[str] = []
        if self.libvirt_domain_driver is LibvirtDomainDriver.KVM:
            required = ["cpus", "ram", "nic_driver", "disk_driver"]
        elif self.libvirt_domain_driver is not LibvirtDomainDriver.NONE:
            required = ["ram"]
        for key in required:
            if not getattr(self, key, None):
                raise ValueError(
                    f"{key} must be specified when libvirt_domain_driver is"
                    f" {self.libvirt_domain_driver.value}"
                )
        return self


PhysicalField = Annotated[str, Field(min_length=1, max_length=32)]

LoopBackField = Annotated[str, Field(min_length=1, max_length=32)]

ManagementField = Annotated[str, Field(min_length=1, max_length=32)]

SERIAL_PORTS_LIMITS = {
    LibvirtDomainDriver.DOCKER: 2,
    LibvirtDomainDriver.IOL: 2,
    LibvirtDomainDriver.KVM: 4,
    LibvirtDomainDriver.NONE: 0,
}


class Interfaces(BaseModel, extra="forbid"):
    """
    Interface configurations.
    """

    serial_ports: int = Field(
        ...,
        description=f"""
            Number of serial ports (console, aux, ...). Maximum value depends on
            simulation libvirt domain driver type: {SERIAL_PORTS_LIMITS}.
        """,
        ge=0,
        le=4,
    )
    default_console: int = Field(
        default=None,
        description="Default serial port for console connections.",
        ge=0,
        le=4,
    )
    physical: conlist(PhysicalField, min_length=1) = Field(
        ..., description="List of physical interfaces."
    )
    has_loopback_zero: bool = Field(
        ..., description="Has `loopback0` interface (used with ANK)."
    )
    min_count: int = Field(
        default=None,
        description="Minimal number of physical interfaces needed to start a node.",
        ge=0,
        le=64,
    )
    default_count: int = Field(default=None, description="Default number of physical interfaces.", ge=1, le=64)
    iol_static_ethernets: Literal[0, 4, 8, 12, 16] = Field(
        default=None,
        description="Only for IOL nodes, the number of static Ethernet interfaces"
        " preceding any serial interface; default 0 means "
        "all interfaces are Ethernet.",
    )
    loopback: conlist(LoopBackField, min_length=1) = Field(default=None, description="List of loopback interfaces.")
    management: conlist(ManagementField, min_length=1) = Field(default=None, description="List of management interfaces.")

    @model_validator(mode="after")
    def validate_one_of(self):
        if self.has_loopback_zero and not self.loopback:
            raise ValueError("loopback must be specified when has_loopback_zero=True")
        return self


class VMProperties(BaseModel, extra="forbid"):
    """Virtual Machine properties."""

    ram: bool = Field(..., description="RAM")
    cpus: bool = Field(..., description="CPU Count.")
    data_volume: bool = Field(..., description="Data Disk Size.")
    boot_disk_size: bool = Field(..., description="Boot Disk Size.")
    cpu_limit: bool = Field(default=None, description="CPU Limit.")


CompletedNode = Annotated[str, Field(max_length=128)]


class Boot(BaseModel, extra="forbid"):
    timeout: int = Field(..., description="Timeout (seconds).", examples=[60], le=86400)
    completed: conlist(CompletedNode, min_length=1) = Field(
        default=None,
        description='A list of strings which should be matched to determine when the node is "ready".',
        examples=[[CompletedNode("string")]],
    )
    uses_regex: bool = Field(
        default=None,
        description="Whether the strings in `completed` should be treated as regular expressions or not.",
    )


class UsageEstimations(BaseModel, extra="forbid"):
    """
    Estimated resource usage.
    """

    cpus: int = Field(
        default=None,
        description=(
            "Estimated CPUs usage in one-hundred-part shares of whole CPUs "
            "(up to 128 CPUs / 12800 shares)."
        ),
        ge=1,
        le=12800,
        examples=[40],
    )
    ram: int = Field(
        default=None,
        description="Estimated RAM usage in MiB.",
        ge=1,
        le=1048576,
        examples=[50],
    )
    disk: int = Field(
        default=None,
        description="Estimated Disk usage in MiB.",
        ge=1,
        le=4194304,
        examples=[500],
    )


class Simulation(BaseModel, extra="forbid"):
    """
    Simulation parameters.
    """

    linux_native: LinuxNative = Field(
        ..., description="Linux native simulation configuration."
    )
    parameters: NodeParameters
    usage_estimations: UsageEstimations = Field(default=None, description="Estimated resource usage parameters.")


class General(BaseModel, extra="forbid"):
    """
    General information for the node type.
    """

    nature: DeviceNature = Field(
        ..., description='The "nature" / kind of the node type defined here.'
    )
    description: str = Field(
        default=None, description="A description of the node type.", max_length=4096
    )
    read_only: bool = Field(
        default=False,
        description="Whether the node definition can be updated and deleted.",
    )


class ConfigurationDriver(StrEnum):
    asav = "asav"
    alpine = "alpine"
    cat9000v = "cat9000v"
    coreos = "coreos"
    csr1000v = "csr1000v"
    desktop = "desktop"
    fmcv = "fmcv"
    ftdv = "ftdv"
    iosv = "iosv"
    iosvl2 = "iosvl2"
    iosxrv = "iosxrv"
    iosxrv9000 = "iosxrv9000"
    nxosv = "nxosv"
    nxosv9000 = "nxosv9000"
    pagent = "pagent"
    sdwan = "sdwan"
    sdwan_edge = "sdwan_edge"
    sdwan_manager = "sdwan_manager"
    server = "server"
    trex = "trex"
    ubuntu = "ubuntu"
    wan_emulator = "wan_emulator"


class ConfigurationMediaType(StrEnum):
    iso = "iso"
    fat = "fat"
    raw = "raw"
    ext4 = "ext4"


class ConfigurationGenerator(BaseModel, extra="forbid"):
    """
    Generator configuration details.
    """

    driver: ConfigurationDriver | None = Field(..., description="Configuration Driver.")


class ConfigurationFile(BaseModel, extra="forbid"):
    """
    Configuration file details.
    """

    editable: bool = Field(..., description="Is the configuration file editable?")
    name: str = Field(
        ...,
        description="The name of the configuration file.",
        min_length=1,
        max_length=64,
    )
    content: NodeConfigurationContent = Field(default=None)


class ConfigurationProvisioning(BaseModel, extra="forbid"):
    """
    Provisioning configuration details.
    """

    files: conlist(ConfigurationFile, min_length=1) = Field(
        ..., description="List of node configuration file objects."
    )
    media_type: ConfigurationMediaType = Field(
        ..., description="The type of the configuration media."
    )
    volume_name: str = Field(
        ...,
        description="The volume name of the configuration media.",
        min_length=1,
        max_length=32,
    )


class Configuration(BaseModel, extra="forbid"):
    """
    Node definition configuration details.
    """

    generator: ConfigurationGenerator = Field(
        ..., description="Generator configuration details."
    )
    provisioning: ConfigurationProvisioning = Field(default=None, description="Provisioning configuration details.")


class Device(BaseModel, extra="forbid"):
    interfaces: Interfaces = Field(...)


class Icon(StrEnum):
    router = "router"
    switch = "switch"
    server = "server"
    host = "host"
    cloud = "cloud"
    firewall = "firewall"
    access_point = "access_point"
    wlc = "wlc"


URI_BASE64_IMAGE_CONTENT_REGEX = re.compile(
    r"^data:image/(png|jpeg|svg\+xml);base64,((?:[A-Za-z0-9+/]{4})*(?:[A-Za-z0-9+/]{4}|"
    r"[A-Za-z0-9+/]{3}=|[A-Za-z0-9+/]{2}={2}))$"
)


def validate_icon_uri(uri: str) -> str:
    """
    Validates if the string is a valid base64 encoded image URI
    containing PNG, JPEG, or SVG.
    """
    if not URI_BASE64_IMAGE_CONTENT_REGEX.fullmatch(uri):
        raise ValueError("Invalid URI content. Check encoded data type")
    return uri


IconDataUri = Annotated[str, AfterValidator(validate_icon_uri)]


class Ui(BaseModel, extra="forbid"):
    """
    UI-related settings for the node type.
    """

    label_prefix: str = Field(
        ...,
        description="The textual prefix for node labels.",
        min_length=1,
        max_length=32,
    )
    icon: Icon | IconDataUri = Field(
        ...,
        description="The icon to use with this node type."
        "Select one from the list or pass custom URI with "
        "png|jpeg|svg\+xml image encoded to base64 in following "
        "format: data:<MIME_TYPE>;base64,<BASE_64_CONTENT>",
    )

    label: str = Field(
        ..., description="The node type label.", min_length=1, max_length=32
    )
    visible: bool = Field(
        ..., description="Determines visibility in the UI for this node type."
    )
    group: Literal["Cisco", "Others"] = Field(default=None, description="Intended to group similar node types (unused).")
    description: str = Field(
        default=None,
        description="The description of the node type (can be Markdown).",
        max_length=4096,
    )


class UiSimplified(Ui, extra="forbid"):
    has_configuration: bool = Field(default=None)
    show_ram: bool = Field(default=None)
    show_cpus: bool = Field(default=None)
    show_cpu_limit: bool = Field(default=None)
    show_data_volume: bool = Field(default=None)
    show_boot_disk_size: bool = Field(default=None)
    has_config_extraction: bool = Field(default=None)


class Inherited(BaseModel, extra="forbid"):
    image: VMProperties = Field(...)
    node: VMProperties = Field(...)


class PyAts(PyAtsCredentials, extra="forbid"):
    os: PyAtsOs = Field(...)
    series: str = Field(
        default=None,
        description="The device series as defined by pyATS / Unicon.",
        min_length=1,
        max_length=32,
    )
    model: PyAtsModel = Field(default=None)
    use_in_testbed: bool = Field(
        default=True, description="Use this device in an exported testbed?"
    )
    config_extract_command: str | None = Field(
        default=None,
        description="This is the CLI command to use when configurations "
        "should be extracted from a device of this node type.",
        max_length=4096,
    )


Schema = Annotated[
    str,
    Field(
        description="The schema version used for this node type.",
        min_length=1,
        max_length=32,
        examples=["0.0.1"],
    ),
]


class NodeDefinition(BaseModel, extra="forbid"):
    id: DefinitionID = Field(
        ...,
        description="""
            A symbolic name used to identify this node definition, such as `iosv` or
            `asav`.
        """,
    )
    boot: Boot = Field(...)
    sim: Simulation = Field(...)
    general: General = Field(...)
    configuration: Configuration = Field(...)
    device: Device = Field(...)
    ui: Ui = Field(...)
    inherited: Inherited = Field(default=None)
    pyats: PyAts = Field(default=None)
    schema_version: Schema = Field(default=None)
    image_definitions: list[DefinitionID] = Field(
        default_factory=list,
        description="""
            The current list of installed image definitions that are associated with the
            node definition.
        """,
    )

    @model_validator(mode="after")
    def validate_serial_consoles_number(self):
        serial_ports = self.device.interfaces.serial_ports
        domain_driver = self.sim.linux_native.libvirt_domain_driver

        max_limit = SERIAL_PORTS_LIMITS.get(domain_driver)

        if serial_ports > max_limit:
            raise ValueError(
                f"{domain_driver} supports up to {serial_ports} serial consoles."
            )

        return self


class SimulationUiSimplified(BaseModel, extra="forbid"):
    """
    Simplified simulation parameters.
    """

    parameters: NodeParameters
    ram: int = Field(default=None, ge=0)
    cpus: int = Field(default=None, ge=0)
    cpu_limit: int = Field(default=100, ge=20, le=100)
    data_volume: int = Field(default=None, ge=0)
    boot_disk_size: int = Field(default=None, ge=0)
    console: bool = Field(default=None)
    simulate: bool = Field(default=None)
    custom_mac: bool = Field(default=None)
    vnc: bool = Field(default=None)


class SimplifiedNodeDefinition(BaseModel, extra="forbid"):
    id: DefinitionID = Field(
        ...,
        description="""
            A symbolic name used to identify this node definition, such as `iosv` or
            `asav`.
        """,
    )
    general: General = Field(...)
    device: Device = Field(...)
    ui: UiSimplified = Field(...)
    sim: SimulationUiSimplified = Field(...)
    image_definitions: list[DefinitionID] = Field(default_factory=list)



#
# This file is part of VIRL 2
# Copyright (c) 2019-2025, Cisco Systems, Inc.
# All rights reserved.
#
import re

from pydantic import BaseModel, Field, model_validator

from simple_webserver.schemas.common import DefinitionID, FilePath
from simple_webserver.schemas.nodes import NodeConfigurationContent
from simple_webserver.schemas.pyats import PyAtsCredentials


class ImageDefinition(BaseModel, extra="forbid"):
    id: DefinitionID = Field(
        ..., description="The identifier of this image definition."
    )
    node_definition_id: DefinitionID = Field(
        ..., description="Node definition ID for the image definition."
    )
    label: str = Field(
        ...,
        description="A required label for the image definition.",
        min_length=1,
        max_length=64,
    )
    disk_image: FilePath | None = Field(
        default=None, description="A source image for the image definition."
    )
    efi_boot: bool = Field(
        default=False,
        description="Whether to use EFI for booting.",
    )
    docker_tag: str | None = Field(
        default=None, description="Docker image tag (e.g. 'alpine:3.19')"
    )
    sha256: str | None = Field(
        default=None,
        description="SHA256 of the disk_image (optional)",
        examples=["58ce6f1271ae1c8a2006ff7d3e54e9874d839f573d8009c20154ad0f2fb0a225"],
        min_length=64,
        max_length=64,
        pattern=re.compile(r"^[a-fA-F\d]{64}(?![\n\r])$"),
    )
    schema_version: str = Field(
        default="0.0.1",
        description="The image definition schema version.",
        examples=["0.0.1"],
        min_length=1,
        max_length=32,
    )
    description: str = Field(
        default="",
        min_length=0,
        max_length=4096,
        description="An optional description for the image definition.",
    )
    disk_image_2: FilePath | None = Field(
        default=None, description="A second source image for the image definition."
    )
    disk_image_3: FilePath | None = Field(
        default=None, description="A third source image for the image definition."
    )
    disk_image_4: FilePath | None = Field(
        default=None, description="A fourth source image for the image definition."
    )
    read_only: bool = Field(
        default=False,
        description="Whether the image definition can be updated or deleted.",
    )
    configuration: NodeConfigurationContent = None
    pyats: PyAtsCredentials | None = None
    ram: int | None = Field(default=None, ge=1, le=1048576, description="Memory (MiB).")
    cpus: int | None = Field(default=None, ge=1, le=128, description="CPUs.")
    cpu_limit: int | None = Field(default=None, ge=20, le=100, description="CPU Limit.")
    data_volume: int | None = Field(
        default=None, ge=0, le=4096, description="Data Disk Size (GiB)."
    )
    boot_disk_size: int | None = Field(
        default=None, ge=0, le=4096, description="Boot Disk Size (GiB)."
    )

    @model_validator(mode="after")
    def check_source(self):
        if self.disk_image or self.docker_tag:
            return self
        raise ValueError("Either disk_image or docker_tag must be provided")



#
# This file is part of VIRL 2
# Copyright (c) 2019-2025, Cisco Systems, Inc.
# All rights reserved.
#
from typing import Annotated

from pydantic import BaseModel, Field

from simple_webserver.schemas.common import UUID4Type


class ErrorResponse(BaseModel, extra="forbid"):
    code: int = Field(
        ..., description="The HTTP status that was associated with this error."
    )
    description: str = Field(
        ..., description="A human-readable message that describes the error."
    )


class FreeFormResponse(BaseModel, extra="forbid"):
    """A freeform JSON object."""

    pass


class IdResponse(BaseModel, extra="forbid"):
    """
    Successful POST / Create operation.
    The response body is a JSON object that indicates the ID of the created object.
    """

    id: UUID4Type = Field(...)


ConvergenceResponse = Annotated[
    bool,
    Field(
        description="The response body is a JSON object with a boolean indicating if convergence has occurred"
    ),
]



#
# This file is part of VIRL 2
# Copyright (c) 2019-2025, Cisco Systems, Inc.
# All rights reserved.
#
from typing import Annotated

from fastapi import Body, Path
from pydantic import BaseModel, Field

from simple_webserver.schemas.common import (
    BaseDBModel,
    GenericDescription,
    GroupName,
    UUID4ArrayType,
    UUID4Type,
)
from simple_webserver.schemas.labs import GroupLab, LabGroupAssociation

GroupIdPathParameter = Annotated[
    UUID4Type, Path(description="The unique ID of a group on this controller.")
]


GroupNamePathParameter = Annotated[
    str,
    Path(
        description="The group name of a group on this controller.",
        examples=["CCNA Study Group Class of 21"],
        min_length=1,
        max_length=64,
    ),
]


class GroupBase(BaseModel):
    """Base class for group create/update."""

    description: GenericDescription = Field(
        default=None,
        description="Additional, textual free-form detail of the group.",
        examples=["CCNA study group"],
    )
    members: UUID4ArrayType = Field(
        default=[],
        description="Members of the group as a list of user IDs.",
    )


class GroupCreateBase(GroupBase):
    """Information needed to create a group."""

    name: GroupName = Field(...)


class GroupUpdateBase(GroupBase):
    """Information needed to update a group."""

    name: GroupName = Field(default=None)


class GroupAssocNew(BaseModel):
    associations: list[LabGroupAssociation] = Field(
        default=None, description="Array of lab/group associations."
    )


class GroupAssocOld(BaseModel):
    labs: list[GroupLab] = Field(
        default=None,
        description="Labs of the group as a object of lab IDs and permission.",
    )


class GroupCreate(GroupCreateBase, GroupAssocNew, extra="forbid"):
    pass


class GroupCreateOld(GroupCreateBase, GroupAssocOld, extra="forbid"):
    pass


class GroupUpdate(GroupUpdateBase, GroupAssocNew, extra="forbid"):
    pass


class GroupUpdateOld(GroupUpdateBase, GroupAssocOld, extra="forbid"):
    pass


GroupCreateBodyParameter = Annotated[GroupCreate | GroupCreateOld, Body(...)]

GroupUpdateBodyParameter = Annotated[GroupUpdate | GroupUpdateOld, Body(...)]


class GroupResponse(
    BaseDBModel, GroupCreateBase, GroupAssocNew, GroupAssocOld, extra="forbid"
):
    """Information about a group."""

    # XXX: Allow this to be None for backward compatibility with existing groups.
    directory_dn: str | None = Field(
        default=None,
        description="Group distinguished name from LDAP",
        max_length=255,
        examples=["CN=Lab 1 Members,CN=groups,DC=corp,DC=com"],
    )
    directory_exists: bool | None = Field(
        default=None, description="Whether the group exists on LDAP"
    )


class GroupBriefResponse(BaseModel, extra="forbid"):
    id: UUID4Type = Field(...)
    name: GroupName = Field(...)
